import itertools
import io
import sys
import numpy as np
from numpy import unravel_index
from numpy import std
from numpy import arange
from numpy import absolute
from numpy import mean
from numpy import std
from array import array
from pandas import read_csv
import pandas as pd
from scipy import stats
from scipy.stats import zscore
from sklearn.model_selection import cross_val_score
from sklearn.model_selection import GridSearchCV
from sklearn.model_selection import KFold
from sklearn.model_selection import RepeatedKFold
from sklearn import linear_model
from sklearn.linear_model import ElasticNet
from sklearn.linear_model import LogisticRegression
from sklearn.metrics import mean_absolute_error
from sklearn.exceptions import ConvergenceWarning
from sklearn import preprocessing
from sklearn import metrics
from sklearn.metrics import roc_auc_score
from sklearn.preprocessing import normalize
from sklearn.metrics import roc_curve
from warnings import filterwarnings
from time import sleep
import warnings
from sklearn.metrics import r2_score
from scipy.stats import pearsonr
from matplotlib import pyplot
import statsmodels.api as sm
from statsmodels.imputation import mice


def close_event():
    pyplot.close() #timer calls this function after 3 seconds and closes the window 

def sprint(*args, **kwargs):
    newstr = ""
    for a in args:
        newstr+=str(a)+''
    return newstr

def pearsonr_ci(x,y,alpha=0.05):
    ''' calculate Pearson correlation along with the confidence interval using scipy and numpy

    Parameters
    ----------
    x, y : iterable object such as a list or np.array
      Input for correlation calculation
    alpha : float
      Significance level. 0.05 by default

    Returns
    -------
    r : float
      Pearson's correlation coefficient
    pval : float
      The corresponding p value
    lo, hi : float
      The lower and upper bound of confidence intervals
    '''

    r, p = stats.pearsonr(x,y)
    r_z = np.arctanh(r)
    se = 1/np.sqrt(x.size-3)
    z = stats.norm.ppf(1-alpha/2)
    lo_z, hi_z = r_z-z*se, r_z+z*se
    lo, hi = np.tanh((lo_z, hi_z))
    return r, p, lo, hi


def bootstrap_auc(clf, X_train, y_train, X_test, y_test, nsamples=1000):
    auc_values = []
    for b in range(nsamples):
        idx = np.random.randint(X_train.shape[0], size=X_train.shape[0])
        clf.fit(X_train[idx], y_train[idx])
        pred = clf.predict_proba(X_test)[:, 1]
        roc_auc = roc_auc_score(y_test.ravel(), pred.ravel())
        auc_values.append(roc_auc)
    return np.percentile(auc_values, (2.5, 97.5))


if (len(sys.argv) == 6):
    DataPath = sys.argv[1]
    LabelFilter = sys.argv[2]
    LogFile = sys.argv[3]
    NSAMPLES = int(sys.argv[4])
    NBestLabelFile=sys.argv[5]
else:
    print('\n\nUsage: ElasticNetClassifier <data> <labelfilter> <logfile> <NSAMPLES> <n-bestlabels>\n\n')
    exit()

print(f'\n\n============================\n\tLabelFilter = {LabelFilter}\n\n')

KFOLD=10

ff = open(LabelFilter, 'r')
LabelFilterList = []
for line in ff:
    LabelFilterList.append(line.strip())
ff.close()


BESTcoeffs={}
resultso = []

Pred_Imputations = []
Corr_Imputations = []
r2_Imputations = []
CorrP_Imputations = []
MAE_Imputations = []


lab = preprocessing.LabelEncoder()

dataframe = read_csv(DataPath, header='infer')

filtereddataframe = dataframe.loc[:,LabelFilterList]


filtereddataframe = filtereddataframe.drop(filtereddataframe[filtereddataframe[LabelFilterList[-1]] == 'exclude'].index)
filtereddataframe[LabelFilterList[-1]] = filtereddataframe[LabelFilterList[-1]].astype(float)    


for col in filtereddataframe.columns[:-1]:
    coldata = filtereddataframe[col]
    if ((coldata[0] == coldata).all()):
        print(f'removing {col}')
        LabelFilterList.remove(col)
        filtereddataframe.drop([col], axis='columns', inplace=True)

nofd = filtereddataframe.loc[:,LabelFilterList[0:-1]]

impute_column_headers = []
for column_headers in nofd.columns:
    coldata=filtereddataframe[column_headers]
    if (coldata.dtype == 'O'):
        impute_column_headers.append(column_headers)


for col in impute_column_headers:
    filtereddataframe[col].replace('impute', np.nan, inplace=True)
    filtereddataframe[col] = filtereddataframe[col].astype(float)


missing_counts = filtereddataframe[impute_column_headers].isna().sum()

print(missing_counts)
print("Total Missing:", missing_counts.sum())

micedataframe = filtereddataframe.reset_index(drop=True)

IMPUTATIONS=1
if (not(missing_counts.empty)):
    mice_data = mice.MICEData(micedataframe[LabelFilterList])
    IMPUTATIONS=10

for IMPUTATION in range(1,IMPUTATIONS+1):
    if (not(missing_counts.empty)):        
        mice_data.update_all()
        micedataframe[impute_column_headers] = mice_data.data[impute_column_headers]
        print(f"Imputation {IMPUTATION}:\n", micedataframe)

    zmicedataframe = micedataframe.copy()

    for col in micedataframe.columns[:-1]:
        zmicedataframe[col] = zscore(micedataframe[col])
        
    data = zmicedataframe.values
        
    X_bss, y_bss = data[:, :-1], lab.fit_transform(data[:, -1])
    
    
    model = LogisticRegression(solver='saga',penalty='elasticnet',multi_class='ovr',class_weight={0: 15},max_iter=10000,tol=0.01,n_jobs=-1)

    Save_fold_Xtest = []
    Save_fold_ytest = []
    Save_fold_ytrain = []
    Save_fold_Pred = []
    Save_fold_BootPred = []
    Save_fold_MAE = []
    Save_fold_regr_Pred = []
    Save_fold_regr_MAE = []
    resultsi = []

    cvo = KFold(n_splits=KFOLD, shuffle=True, random_state=42)    
    for train_index_o, test_index_o in cvo.split(X_bss,y=y_bss):
        X_train_o, X_test_o = X_bss[train_index_o], X_bss[test_index_o]        
        y_train_o, y_test_o = y_bss[train_index_o], y_bss[test_index_o]
        
        for col in arange(X_train_o.shape[1]):
            mean_X_train_o = mean(X_train_o[:,col])
            X_train_o[:,col] = (X_train_o[:,col] - mean_X_train_o)
            X_test_o[:,col] = (X_test_o[:,col] - mean_X_train_o)
        
        # define grid
        grid = dict()
        grid['alpha'] = np.logspace(-4, 4, 9)
        grid['l1_ratio'] = arange(0, 1.01, 0.01)
        grid['score'] = np.zeros((grid['alpha'].shape[0],grid['l1_ratio'].shape[0]),dtype=float)
        
        alfai = -1
        for alfa in grid['alpha']:
            alfai = alfai+1
            l1ri = -1
            for l1r in grid['l1_ratio']:
                l1ri =  l1ri+1
                modeli = LogisticRegression(solver='saga',penalty='elasticnet',C=alfa,l1_ratio=l1r,multi_class='ovr',class_weight={0: 15},max_iter=10000,tol=0.01,n_jobs=-1)
                
                grid['score'][alfai][l1ri]=0
                
                cvi = KFold(n_splits=KFOLD, shuffle=True, random_state=42)    
                for train_index_i, test_index_i in cvi.split(X_train_o,y=y_train_o):                    
                    X_train_i, X_test_i = X_train_o[train_index_i], X_train_o[test_index_i]        
                    y_train_i, y_test_i = y_train_o[train_index_i], y_train_o[test_index_i]

        
                    for col in arange(X_train_i.shape[1]):
                        mean_X_train_i = mean(X_train_i[:,col])
                        X_train_i[:,col] = (X_train_i[:,col] - mean_X_train_i)
                        X_test_i[:,col] = (X_test_i[:,col] - mean_X_train_i)
                        
                    modeli.fit(X_train_i,y_train_i)

                    yhat = modeli.predict_proba(X_test_i)

                    grid['score'][alfai][l1ri] = grid['score'][alfai][l1ri] + (mean_absolute_error(y_test_i,yhat[:,1])**2)
                    
        err=float('NaN')
        while(np.isnan(err)):
            min_index = unravel_index(grid['score'].argmin(), grid['score'].shape)
            grid['score'][min_index[0]][min_index[1]]=10000000000
            best_alfa=grid['alpha'][min_index[0]]
            best_l1r=grid['l1_ratio'][min_index[1]]

            modeli = LogisticRegression(solver='saga',penalty='elasticnet',C=best_alfa,l1_ratio=best_l1r,multi_class='ovr',class_weight={0: 15},max_iter=10000,tol=0.01,n_jobs=-1)
            modeli.fit(X_train_o,y_train_o)
            
            yhat = modeli.predict_proba(X_test_o)
            
            err = mean_absolute_error(y_test_o,yhat[:,1])

            
        print(f'alpha = {best_alfa} , l1_ratio = {best_l1r}')

        MAE = mean_absolute_error(y_test_o,yhat[:,1])

        ytest = np.empty((NSAMPLES,y_test_o.shape[0]))
        pred = np.empty((NSAMPLES,y_test_o.shape[0]))
        for b in range(NSAMPLES):
            idx = np.random.randint(X_train_o.shape[0], size=X_train_o.shape[0])
            modeli.fit(X_train_o[idx], y_train_o[idx])
            pred[b] = modeli.predict_proba(X_test_o)[:, 1]
        
    
        Save_fold_Xtest.append(X_test_o)
        Save_fold_ytrain.append(y_train_o)
        Save_fold_ytest.append(y_test_o)
        Save_fold_Pred.append(yhat[:,1])
        Save_fold_BootPred.append(pred)
        Save_fold_MAE.append(MAE)
    
        print('MAE: %.3f' % MAE)

        values = np.concatenate(abs(modeli.coef_))
        print(f'maxvalue = {values[np.argmax(values)]}')
                
        for k in arange(1,len(list(filtereddataframe.columns.values))):
            maxcoefindex = np.argmax(values)
            column_names = list(filtereddataframe.columns.values)
            if (values[maxcoefindex] > 0):
                if column_names[maxcoefindex] in BESTcoeffs.keys():
                    keyval =  BESTcoeffs[column_names[maxcoefindex]]
                    newCnt = keyval[1] + 1
                    newVal = max( keyval[0] , values[maxcoefindex] )
                    BESTcoeffs[column_names[maxcoefindex]] = [ newVal , newCnt ]                  
                else:
                    BESTcoeffs[column_names[maxcoefindex]] = [ values[maxcoefindex] , 1 ]
            values[maxcoefindex] = -1
        
    print(f'---------------\n')


    # save the NBEST best coeffs
    NBestLabelFile1 = NBestLabelFile
    if (not(missing_counts.empty)):        
        NBestLabelFile1 = NBestLabelFile.replace(".best", "")
        NBestLabelFile1 = sprint(f'{NBestLabelFile1}_IMPUTATION{IMPUTATION}.best')
        
    fb = open(NBestLabelFile1, 'w', encoding=None, buffering=1)

    maxc = 0
    for k in BESTcoeffs.keys():
        c = BESTcoeffs[k][1]
        maxc = max(maxc,c)

    most_BESTcoeffs = dict((k,v[0]) for k, v in BESTcoeffs.items() if v[1] == maxc)
    sorted_BESTcoeffs = dict(sorted(most_BESTcoeffs.items(), key=lambda x: x[1], reverse=True))
    
    for x, y in sorted_BESTcoeffs.items():
        s = sprint(f'{x} : {y}\n')
        fb.write(s)

    fb.close()

    
    pred_boot = [ [Save_fold_BootPred[0][i]] for i in arange(NSAMPLES)]
    for i in arange(NSAMPLES):
        for j in arange(1,KFOLD):
            pred_boot[i].append(Save_fold_BootPred[j][i])
        pred_boot[i] = np.concatenate(pred_boot[i])
    
    Pred_Boots = np.vstack((pred_boot[:]))
    
    All_Xtest = np.concatenate(Save_fold_Xtest)
    All_ytest = np.concatenate(Save_fold_ytest)
    All_ytrain = np.concatenate(Save_fold_ytrain)
    All_Pred = np.concatenate(Save_fold_Pred)

    if (IMPUTATION==1):
        Pred_Imputations = np.reshape(All_Pred, (1,All_Pred.shape[0]))
        ytest_Imputations = np.reshape(All_ytest, (1,All_ytest.shape[0]))
        PredBoots_Imputations = np.reshape(Pred_Boots,(1,Pred_Boots.shape[0],Pred_Boots.shape[1]))
    else:
        Pred_Imputations = np.vstack((Pred_Imputations,All_Pred))
        ytest_Imputations = np.vstack((ytest_Imputations,All_ytest))
        PredBoots_Imputations = np.vstack((PredBoots_Imputations,np.reshape(Pred_Boots,(1,Pred_Boots.shape[0],Pred_Boots.shape[1]))))
    
    LogFile1 = LogFile
    if (not(missing_counts.empty)):        
        LogFile1 = LogFile.replace(".log", "")
        LogFile1 = sprint(f'{LogFile1}_IMPUTATION{IMPUTATION}.log')
    
    f = open(LogFile1, 'w', encoding=None, buffering=1)
    s = sprint(f'DataPath = {DataPath}\n')
    f.write(s)
    print(f'DataPath = {DataPath}\n')

    s = sprint(f'LabelFilter = {LabelFilter}\n')
    f.write(s)
    print(f'LabelFilter = {LabelFilter}\n')
        
    Corr_all = pearsonr(All_ytest, All_Pred)[0]
    Corr_Imputations.append(Corr_all)

    print("Correlation = ",Corr_all)
    s = sprint("Correlation = ",Corr_all)
    f.write(s+'\n')

    
    Corr_all_p = pearsonr(All_ytest, All_Pred)[1]
    CorrP_Imputations.append(Corr_all_p)
    
    print("Correlation p-value = ",Corr_all_p)
    s = sprint("Correlation p-value = ",Corr_all_p)
    f.write(s+'\n')
    
    r2_all = r2_score(All_ytest, All_Pred)
    r2_Imputations.append(r2_all)

    
    print("r2 = ", r2_all)
    s = sprint("r2 = ", r2_all)
    f.write(s+'\n')
    
    MAE_all =mean_absolute_error(All_ytest, All_Pred)    
    MAE_Imputations.append(MAE_all)

    MAE_all =mean_absolute_error(All_ytest, All_Pred)
    print("Mean absolute error = ",MAE_all)
    s = sprint("Mean absolute error = ",MAE_all)
    f.write(s+'\n')
    
    AUC_all = metrics.roc_auc_score(All_ytest, All_Pred)
    print(f'AUC_all = {AUC_all}')
    s = sprint(f'AUC_all = {AUC_all}\n')
    f.write(s)


    f.close()

# ====================================================================================================


LogFile1 = LogFile
if (not(missing_counts.empty)):        
    LogFile1 = LogFile.replace(".log", "")
    LogFile1 = sprint(f'{LogFile1}_imp{IMPUTATIONS}.log')



f = open(LogFile1, 'w', encoding=None, buffering=1)
s = sprint(f'DataPath = {DataPath}\n')
f.write(s)
print(f'DataPath = {DataPath}\n')

s = sprint(f'LabelFilter = {LabelFilter}\n')
f.write(s)
print(f'LabelFilter = {LabelFilter}\n')
        
# calculate the AUC for each imputation
PredBoots_All = np.reshape(PredBoots_Imputations,(PredBoots_Imputations.shape[0]*PredBoots_Imputations.shape[1], PredBoots_Imputations.shape[2]))
auc_values=[]
for i in arange(PredBoots_All.shape[0]):
    roc_auc = roc_auc_score(All_ytest, PredBoots_All[i])
    auc_values.append(roc_auc)

SaveFile = LogFile.replace(".log", "")
for i in arange(PredBoots_Imputations.shape[0]):
    s=sprint(f'{SaveFile}_PredBoots_Imputations[{i}].txt')
    np.savetxt(s, np.array(PredBoots_Imputations[i]))

All_ytest = np.concatenate(Save_fold_ytest)
s=sprint(f'{SaveFile}_ytest.txt')
np.savetxt(s, All_ytest)
    
auc_cilo, auc_mid, auc_cihi = np.percentile(auc_values, (2.5, 50.0, 97.5))    

print(f'loauc={auc_cilo} . auc_mid={auc_mid} , hiauc={auc_cihi}')

All_Pred = np.concatenate(Pred_Imputations)
All_ytest = np.concatenate(ytest_Imputations)


AUC_all = metrics.roc_auc_score(All_ytest, All_Pred)
print(f'AUC_all = {AUC_all}')
s = sprint(f'AUC_all = {AUC_all}\n')
f.write(s)


fpr1, tpr1, thresh1 = roc_curve(All_ytest, All_Pred, pos_label=1)


fig, (ax1, ax2) = pyplot.subplots(1, 2)
fig.set_size_inches(14,7)

ax1.plot(fpr1, tpr1, linestyle='--',color='orange', label='ElasticNet Regression')


textfont = {'family': 'serif',
            'color':  'darkred',
            'weight': 'bold',
            'size': 10,
            }

ax1.axes.set_xlabel('False Positive Rate', textfont)
ax1.axes.set_ylabel('True Positive Rate', textfont)

t = sprint(f' {LabelFilterList[-1]}\n')
titlefont = {'family': 'serif',
             'color':  'darkred',
             'weight': 'bold',
             'size': 12,
             }

ax1.set_title(t,fontdict=titlefont)


ittextfont = {'family': 'serif',
              'color':  'darkred',
              'style': 'italic',
              'weight': 'bold',
              'size': 10,
              }

left, width = .25, .725
hmiddle = left + width/2
bottom, height = .025, .1
right = left + width
top = bottom + height
vmiddle1 = bottom + height*0.66
vmiddle2 = bottom + height*0.33
p = pyplot.Rectangle((left, bottom), width, height, linewidth=0, fill=False)
p.set_transform(ax1.transAxes)
p.set_clip_on(False)
ax1.add_patch(p)

s = sprint(f'AUC ')

ax1.text(hmiddle, top, s,
         fontdict=ittextfont,
         horizontalalignment='right',
         verticalalignment='center',
         transform=ax1.transAxes)

s = sprint(f'= %.3f' % auc_mid)
s = sprint(f'{s} [%.3f' % auc_cilo)    
s = sprint(f'{s} , %.3f]' % auc_cihi)    

ax1.text(hmiddle, top, s,
         fontdict=textfont,
         horizontalalignment='left',
         verticalalignment='center',
         transform=ax1.transAxes)

s = sprint(f'p ')


s = sprint(f'= %.3f' % mean(CorrP_Imputations))    

s = sprint(f'r$^2$ ')


s = sprint(f'= %.3f'% mean(r2_Imputations))


s = sprint(f'MAE ')


s = sprint(f'= %.3f'% mean(MAE_Imputations))


# -----------------------

ShortLabelFilterList = []
for i in arange(len(LabelFilterList)-1):
    splits = str.split(LabelFilterList[i],'_')
    if (ShortLabelFilterList.count(splits[0])==0):
        ShortLabelFilterList.append(splits[0])

fpic = LogFile.replace(".log", "")            
if (not(missing_counts.empty)):
    fpic = sprint(f'{fpic}_imp{IMPUTATIONS}.png')
else:
    fpic = sprint(f'{fpic}.png')        

print(f'{fpic}\n')

sorted_BEST25coeffs = dict(itertools.islice(sorted_BESTcoeffs.items(),25))

X=arange(len(list(sorted_BEST25coeffs.values()))).tolist()
Y = list(sorted_BEST25coeffs.values())
newY1 = [e+1 for e in Y]


ax2.scatter(X,Y)


ax2.spines[['right', 'top']].set_visible(False)

ax2.set_xlim([-0.5, len(sorted_BEST25coeffs)-0.01])
if (len(Y)>0):
    ax2.set_ylim([0, Y[0]*2])

t = sprint(f'Top {len(sorted_BEST25coeffs)} predictors\n')
ax2.axes.set_xlabel(t, textfont)
ax2.axes.set_ylabel('Coefficients', textfont)

labels=[]
for i in arange(len(sorted_BEST25coeffs)):
    s=sprint(f'{i+1}')
    labels.append(s)

ax2.set_xticks(np.arange(len(sorted_BEST25coeffs)),labels=labels)

keyfont = {'family': 'serif',
           'color':  'darkred',
           'weight': 'bold',
           'size': 6,
           'rotation' : 'vertical'
           }

for i, key in enumerate(sorted_BEST25coeffs.keys()):
    val=sorted_BEST25coeffs[key]
    ax2.text(i, val+(Y[0]*0.05), key, fontdict=keyfont)

 
pyplot.savefig(fpic)
    
timer = fig.canvas.new_timer(interval = 30000) 
timer.add_callback(close_event)


timer.start()
pyplot.show()

print(f'\n\n============================\n\n')

f.close()
