"""

# I guess I read in the csv data, each line of which specifies a t1, t2, SDM, and target.
# I then read in that data, and turn it into a dataloader ... if I understand that correctly.
# I then use the dataloader to deliver randomized minibatches of data during training.
#
# A question that I need to deal with is how to normalize the data.  I mean how to make the SDM weigh in to the same extent as the t1 and t2.  Or perhaps I don't care ; the t1 and t2 are both normalized in the 0 to 100 range, and the SDM is something like -5 to 25.  Does that matter?
#
# I will run this with one argument: the csv files that lists all of the data. 
 
"""
import numpy as np
from numpy import genfromtxt
import pandas as pd
import nibabel as nib
import re
import glob
import os
import scipy.stats as stats
from scipy.stats import linregress
from sklearn.model_selection import KFold
from sklearn.metrics import mean_absolute_error
from sklearn.model_selection import LeaveOneOut
import sys
import torch
from torch.utils.data import Dataset
from torch.utils.data import DataLoader
from torchvision import datasets
from torchvision import transforms    
import time
from torch import optim
from torch.nn.functional import normalize
# 3D-UNet model.
# x: 128x128 resolution for 32 frames.
import torch.nn as nn
from torchmetrics.functional import mean_absolute_error



class BGTDataset(Dataset):
    """Labelled BGT dataset."""

    def __init__(self, csv_file, root_dir, transform=None):
        """
        Arguments:
            csv_file (string): Path to the csv file listing t1, t2, SDM, and target label files.
            root_dir (string): Directory with all the images.
            transform (callable, optional): Optional transform to be applied
                on a sample.
        """
        self.files = pd.read_csv(csv_file,header=None,delimiter=',')
        self.root_dir = root_dir
        self.transform = transform        
        
    def __len__(self):
        return len(self.files)

    def __getitem__(self, idx):    

        t1_name = self.files.iat[idx,0]
        #t1_filename=os.path.basename(t1_name).split('/')[-1]
        #substri = t1_filename.find('.nii.gz')
        #t1_basename = t1_filename[0:substri]
        #substri = t1_name.find(t1_filename)
        #t1_pathname=t1_name[0:substri]        
        t1_img = nib.load(self.files.iat[idx,0])
        t1_data = torch.from_numpy(t1_img.get_fdata())
        t1_data = NormalizeData(t1_data)
        
        t2_name = self.files.iat[idx,1]
        #t2_filename=os.path.basename(t2_name).split('/')[-1]
        #substri = t2_filename.find('.nii.gz')
        #t2_basename = t2_filename[0:substri]
        #substri = t2_name.find(t2_filename)
        #t2_pathname=t2_name[0:substri] 
        t2_img = nib.load(self.files.iat[idx,1])
        t2_data = torch.from_numpy(t2_img.get_fdata())
        t2_data = NormalizeData(t2_data)        

        # print(f"t1_name: {t1_name}")

        #for i in range(len(t1_basename)):
        #    if (t1_basename[0:i] == t2_basename[0:i]):
        #        subj_id = t1_basename[0:i]
            
        #substrsj = t1_name.find('.nii.gz')
        #t1_name_tail=t1_name[substrsi:substrsj]
        #substrsij = t1_name_tail.find('/')
        #subj_id=t1_name[substrsi:substrsi+substrsij]
        # print(f"subj_num: {subj_num}")
        

        x_data = torch.stack([t1_data, t2_data])
        x_data = x_data.type(torch.float32)
        x_data = x_data.to(device)


        y_name = self.files.iat[idx,2]
        y_targetname = y_name+'_target.nii.gz'
        y_img = nib.load(y_targetname)
        y_actualdata = torch.from_numpy(y_img.get_fdata())
        # y_data = NormalizeData(y_data)
        # make it binary
        y_data1 = y_actualdata<0.5
        y_data2 = y_actualdata>0.5
        y_data = y_data2
        # y_data = torch.stack([y_data1,y_data2])
        # y_data = torch.stack([y_data])
        # y_data = y_data.float()        
        y_data = y_data.type(torch.int64)
        # y_data = y_data.type(torch.float32)
        y_data = y_data.to(device)
                        
        return x_data, y_data, y_name
  


def accuracy(predictions, labels):
    classes = torch.argmax(predictions, dim=1)
    return torch.mean((classes == labels).float())



def NormalizeData(data):
    return (data - data.min()) / (data.max() - data.min())


def conv_block_3d(in_dim, out_dim, activation):
    return nn.Sequential(
        nn.Conv3d(in_dim, out_dim, kernel_size=3, stride=1, padding=1),
        nn.BatchNorm3d(out_dim),
        activation,)


def conv_trans_block_3d(in_dim, out_dim, activation):
    return nn.Sequential(
        nn.ConvTranspose3d(in_dim, out_dim, kernel_size=3, stride=2, padding=1, output_padding=1),
        nn.BatchNorm3d(out_dim),
        activation,)


def max_pooling_3d():
    return nn.MaxPool3d(kernel_size=2, stride=2, padding=0)


def max_unpooling_3d():
    return nn.MaxUnpool3d(kernel_size=2, stride=2, padding=0)


def conv_block_2_3d(in_dim, out_dim, activation):
    return nn.Sequential(
        conv_block_3d(in_dim, out_dim, activation),
        nn.Conv3d(out_dim, out_dim, kernel_size=3, stride=1, padding=1),
        nn.BatchNorm3d(out_dim),
        activation,)        


class UNet(nn.Module):
    def __init__(self, in_dim, out_dim, num_filters):
        super(UNet, self).__init__()
        
        self.in_dim = in_dim
        self.out_dim = out_dim
        self.num_filters = num_filters
        activation = nn.LeakyReLU(0.2, inplace=True)
        
        # Down sampling
        # self.pool_1a = max_pooling_3d()
        self.down_1 = conv_block_2_3d(self.in_dim, self.num_filters, activation)
        self.pool_1 = max_pooling_3d()
        #        self.pool_2a = max_pooling_3d()
        self.down_2 = conv_block_2_3d(self.num_filters, self.num_filters * 2, activation)
        self.pool_2 = max_pooling_3d()
        #        self.pool_3a = max_pooling_3d()
        self.down_3 = conv_block_2_3d(self.num_filters * 2, self.num_filters * 4, activation)
        self.pool_3 = max_pooling_3d()
        #        self.pool_4a = max_pooling_3d()
        self.down_4 = conv_block_2_3d(self.num_filters * 4, self.num_filters * 8, activation)
        self.pool_4 = max_pooling_3d()
        #        self.pool_5a = max_pooling_3d()
        self.down_5 = conv_block_2_3d(self.num_filters * 8, self.num_filters * 16, activation)
        self.pool_5 = max_pooling_3d()
        
        # Bridge
        #self.bridge = conv_block_2_3d(self.num_filters * 8, self.num_filters * 16, activation)
        self.bridge = conv_block_2_3d(self.num_filters * 16, self.num_filters * 32, activation)
        # self.bridge = conv_block_2_3d(self.num_filters * 8, self.num_filters * 16, activation)
        
        # Up sampling
        # self.up_1a = max_unpooling_3d()
        # self.trans_1 = conv_trans_block_3d(self.num_filters * 4, self.num_filters * 4, activation)
        # self.up_1 = conv_block_2_3d(self.num_filters * 6, self.num_filters * 2, activation)
        self.trans_1 = conv_trans_block_3d(self.num_filters * 32, self.num_filters * 32, activation)
        self.up_1 = conv_block_2_3d(self.num_filters * 48, self.num_filters * 16, activation)
        #        self.up_2a = max_unpooling_3d()
        self.trans_2 = conv_trans_block_3d(self.num_filters * 16, self.num_filters * 16, activation)
        self.up_2 = conv_block_2_3d(self.num_filters * 24, self.num_filters * 8, activation)
        #        self.up_3a = max_unpooling_3d()
        self.trans_3 = conv_trans_block_3d(self.num_filters * 8, self.num_filters * 8, activation)
        self.up_3 = conv_block_2_3d(self.num_filters * 12, self.num_filters * 4, activation)
        #        self.up_4a = max_unpooling_3d()
        self.trans_4 = conv_trans_block_3d(self.num_filters * 4, self.num_filters * 4, activation)
        self.up_4 = conv_block_2_3d(self.num_filters * 6, self.num_filters * 2, activation)
        #        self.up_5a = max_unpooling_3d()
        #        self.trans_5 = conv_trans_block_3d(self.num_filters * 2, self.num_filters * 2, activation)
        #        self.up_5 = conv_block_2_3d(self.num_filters * 3, self.num_filters * 1, activation)
        
        self.trans_5 = conv_trans_block_3d(self.num_filters * 2, self.num_filters * 2, activation)
        self.up_5 = conv_block_2_3d(self.num_filters * 3, self.num_filters * 1, activation)
        # self.unpool_5 = max_unpooling_3d()
        
        # Output
        self.out = conv_block_3d(self.num_filters, out_dim, activation)
    
    def forward(self, x):
        # Down sampling
        # pool_1a = self.pool_1(x) 
        down_1 = self.down_1(x) 
        # down_1 = self.down_1(pool_1a) 
        pool_1 = self.pool_1(down_1) 
        
        #        pool_2a = self.pool_2(pool_1) # -> [1, 8, 16, 16, 16]
        down_2 = self.down_2(pool_1) # -> [1, 8, 32, 32, 32]
        pool_2 = self.pool_2(down_2) # -> [1, 8, 16, 16, 16]
        
        #        pool_3a = self.pool_3(pool_2) # -> [1, 16, 8, 8, 8]
        down_3 = self.down_3(pool_2) # -> [1, 16, 16, 16, 16]
        pool_3 = self.pool_3(down_3) # -> [1, 16, 8, 8, 8]
        
        #        pool_4a = self.pool_4(pool_3) # -> [1, 32, 4, 4, 4]
        down_4 = self.down_4(pool_3) # -> [1, 32, 8, 8, 8]
        pool_4 = self.pool_4(down_4) # -> [1, 32, 4, 4, 4]
        
        #        pool_5a = self.pool_5(pool_4) # -> [1, 64, 2, 2, 2]
        down_5 = self.down_5(pool_4) # -> [1, 64, 4, 4, 4]
        pool_5 = self.pool_5(down_5) # -> [1, 64, 2, 2, 2]
        
        # Bridge
        bridge = self.bridge(pool_5) 
        
        # Up sampling
        # up_1a = self.up_1(bridge)
        trans_1 = self.trans_1(bridge) 
        concat_1 = torch.cat([trans_1, down_5], dim=1)
        up_1 = self.up_1(concat_1)

        #        up_2a = self.up_1(up_1)        
        trans_2 = self.trans_2(up_1)
        concat_2 = torch.cat([trans_2, down_4], dim=1)
        up_2 = self.up_2(concat_2)
        
        #        up_3a = self.up_1(up_2)                
        trans_3 = self.trans_3(up_2)
        concat_3 = torch.cat([trans_3, down_3], dim=1)
        up_3 = self.up_3(concat_3)
        
        #        up_4a = self.up_1(up_3)                        
        trans_4 = self.trans_4(up_3)
        concat_4 = torch.cat([trans_4, down_2], dim=1)
        up_4 = self.up_4(concat_4)
        
        #        up_5a = self.up_1(up_4)                                
        #        trans_5 = self.trans_5(up_5a)
        trans_5 = self.trans_5(up_4)
        concat_5 = torch.cat([trans_5, down_1], dim=1)
        up_5 = self.up_5(concat_5)
        # unpool_5 = self.unpool_5(up_5)
        
        # Output
        # out = self.out(unpool_5)
        out = self.out(up_5)
        softmax = nn.LogSoftmax(dim=1)
        out = softmax(out)
        out = out.type(torch.float32)
        return out



    

if (len(sys.argv) != 2):
    print('\n\nUsage: python BET-CNN.py file.csv\n\n')
    exit()
    
# print('\n\nGot four arguments.  Testing data: '+sys.argv[1]+'\n\n')
# time.sleep(1)

# sys.argv[1]   # This is the augmentation set that will be the testing data
HOLDOUT='0'

# device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
device = torch.device('cpu')
# device = torch.device('cpu' if torch.cuda.is_available() else 'cpu')
# print('Using device:', device)
# print()



# Training_dataset = BGTDataset(csv_file=sys.argv[2], root_dir='')
Testing_dataset = BGTDataset(csv_file=sys.argv[1], root_dir='')
# print('Training dataset length = '+str(len(Training_dataset)))
# print('Testing dataset length = '+str(len(Testing_dataset)))


model = UNet(in_dim=2, out_dim=2, num_filters=8)
#start_epoch = 1



files = glob.glob('/mnt/md0/Academia/Research/Projects/HIE/Cohort1/BET-CNN/BET-5layer8filter_model_'+HOLDOUT+'_*.pt')
# print('================================')
# print(files)
# print('================================')
most_recent_file = None
most_recent_time = 0
if (len(files)==1):
    most_recent_file = files[0]
else:
    for  i in range(len(files)):
        filemtime =  os.path.getmtime(files[i])
        # print('filemtime='+str(filemtime))
        if (filemtime > most_recent_time):
            most_recent_time = filemtime
            most_recent_file=files[i]
            # print('most_recent_file = '+most_recent_file+'\n')
    # print('ERROR: File not found\n')
    # exit()

    
if (most_recent_file == None):
    print('\n\nERROR: missing weight file\n\n')
else:
    # I should extrat the start_epoch from the name
    print('\nLoading  "'+most_recent_file+'"\n')
    time.sleep(1)
    model = torch.load(most_recent_file)
    substrsi = most_recent_file.find('BET-5layer8filter_model_'+HOLDOUT+'_')+len('BET-5layer8filter_model_'+HOLDOUT+'_')
    substrsj = most_recent_file.find('.pt')
    start_epoch = int(most_recent_file[substrsi:substrsj])+1
    
model = model.to(device)
# print(model)
# loss_fn = nn.CrossEntropyLoss(reduction='none')
# loss_fn = nn.BCELoss()
loss_fn = nn.NLLLoss()


# SGD Optimizer
learning_rate = 0.01
optimizer = torch.optim.SGD(model.parameters(), lr=learning_rate)



# for train_index, test_index in loo.split(BGT_dataset):



end_epoch=start_epoch - 1
n_epochs=end_epoch - start_epoch + 1
# mean_epoch_loss = [0] * (n_epochs + 1)
#end_epoch = start_epoch + n_epochs + 1
epoch = start_epoch

with torch.no_grad():

    total_accuracies = 0
    total_mae = 0
    
    batch_size=1
    
    for i in range(0, len(Testing_dataset),batch_size):

        test_dataloader = DataLoader(Testing_dataset, batch_size=1, shuffle=True)        
        test_data, test_targets, test_filenames = next(iter(test_dataloader))

        y_pred = model(test_data)
        y_pred = y_pred.to(device)
        
        loss = loss_fn(y_pred, test_targets)
        # print(f"loss.mean ({test_subjs[0]}) :  {loss.mean()}")
        # ft.write('loss.mean ('+str(end_epoch)+','+test_subjs[0]+') :  '+str(loss.mean().to('cpu').detach().numpy())+'\n')        
        
        prediction = torch.exp(y_pred[0]) 
        prediction = prediction[1] - prediction[0]
        prediction = (prediction>0.5).float()

        # nondistortedAccuracy=accuracy(prediction,test_targets[0])
        # total_accuracies += nondistortedAccuracy
        # print(f"accuracy ({test_subjs[0]}) :  {nondistortedAccuracy}")
        # ft.write('accuracy ('+str(end_epoch)+','+test_subjs[0]+') :  '+str(nondistortedAccuracy.to('cpu').detach().numpy())+'\n')
        
        # mae = mean_absolute_error(prediction,test_targets[0])
        # total_mae += mae
        # print(f"MAE ({test_subjs[0]}) :  {mae}")
        # ft.write('MAE ('+str(end_epoch)+','+test_subjs[0]+') :  '+str(mae.to('cpu').detach().numpy())+'\n')

        print('Saving t1')
        t1_numpy = test_data[0][0].to('cpu').detach().numpy()
        ni_img = nib.Nifti1Image(t1_numpy,affine=np.eye(4),dtype=float)
        nib.save(ni_img, test_filenames[0]+'_bett1.nii')

        # t2_numpy = test_data[0][1].to('cpu').detach().numpy()
        # ni_img = nib.Nifti1Image(t2_numpy,affine=np.eye(4),dtype=float)
        # nib.save(ni_img, 'BET-5layer8filter_model_t2_'+HOLDOUT+'_'+str(epoch-1)+'_'+test_subjs[0]+'.nii')
        
        #target_numpy = test_targets[0].to('cpu').detach().numpy()
        #ni_img = nib.Nifti1Image(target_numpy,affine=np.eye(4),dtype=float)
        # nib.save(ni_img, 'BET-5layer8filter_model_target_'+HOLDOUT+'_'+str(epoch-1)+'_'+test_subjs[0]+'.nii')

        print('Saving prediction')        
        ni_img = nib.Nifti1Image(prediction.to('cpu').detach().numpy(),affine=np.eye(4),dtype=float)
        nib.save(ni_img, test_filenames[0]+'_betprediction.nii')

    # ave_accuracies = total_accuracies/len(Testing_dataset)
    # print(f"average accuracies:  {ave_accuracies}")
    # ft.write('average_accuracies  ('+str(end_epoch)+') :  '+str(ave_accuracies.to('cpu').detach().numpy())+'\n')
    # ft.write('average_accuracies  ('+str(end_epoch)+') :  '+str(ave_accuracies)+'\n')
    
    # ave_mae = total_mae/len(Testing_dataset)
    # print(f"average MAE:  {ave_mae}")
    # ft.write('average MAE  ('+str(end_epoch)+') :  '+str(ave_mae.to('cpu').detach().numpy())+'\n')
    # ft.write('average MAE  ('+str(end_epoch)+') :  '+str(ave_mae)+'\n')

# ft.close()
